@@ -13,7 +13,7 @@ import yaml
 
 from shot_scraper.utils import filename_for_url, url_or_file_path
 
-BROWSERS = ("chromium", "firefox", "chrome", "chrome-beta")
+BROWSERS = ("chromium", "firefox", "webkit", "chrome", "chrome-beta")
 
 
 def browser_option(fn):
@@ -27,6 +27,20 @@ def browser_option(fn):
     return fn
 
 
+def user_agent_option(fn):
+    click.option("--user-agent", help="User-Agent header to use")(fn)
+    return fn
+
+
+def reduced_motion_option(fn):
+    click.option(
+        "--reduced-motion",
+        is_flag=True,
+        help="Emulate 'prefers-reduced-motion' media feature",
+    )(fn)
+    return fn
+
+
 @click.group(
     cls=DefaultGroup,
     default="shot",
@@ -72,6 +86,24 @@ def cli():
     help="Take shot of first element matching this CSS selector",
     multiple=True,
 )
+@click.option(
+    "selectors_all",
+    "--selector-all",
+    help="Take shot of all elements matching this CSS selector",
+    multiple=True,
+)
+@click.option(
+    "js_selectors",
+    "--js-selector",
+    help="Take shot of first element matching this JS (el) expression",
+    multiple=True,
+)
+@click.option(
+    "js_selectors_all",
+    "--js-selector-all",
+    help="Take shot of all elements matching this JS (el) expression",
+    multiple=True,
+)
 @click.option(
     "-p",
     "--padding",
@@ -102,6 +134,8 @@ def cli():
     help="Interact mode with developer tools",
 )
 @browser_option
+@user_agent_option
+@reduced_motion_option
 def shot(
     url,
     auth,
@@ -109,6 +143,9 @@ def shot(
     width,
     height,
     selectors,
+    selectors_all,
+    js_selectors,
+    js_selectors_all,
     padding,
     javascript,
     retina,
@@ -118,6 +155,8 @@ def shot(
     interactive,
     devtools,
     browser,
+    user_agent,
+    reduced_motion,
 ):
     """
     Take a single screenshot of a page or portion of a page.
@@ -146,10 +185,14 @@ def shot(
         shot-scraper https://simonwillison.net -s '#bighead'
     """
     if output is None:
-        output = filename_for_url(url, file_exists=os.path.exists)
+        ext = "jpg" if quality else None
+        output = filename_for_url(url, ext=ext, file_exists=os.path.exists)
     shot = {
         "url": url,
         "selectors": selectors,
+        "selectors_all": selectors_all,
+        "js_selectors": js_selectors,
+        "js_selectors_all": js_selectors_all,
         "javascript": javascript,
         "width": width,
         "height": height,
@@ -169,7 +212,9 @@ def shot(
             devtools=devtools,
             retina=retina,
             browser=browser,
+            user_agent=user_agent,
             timeout=timeout,
+            reduced_motion=reduced_motion,
         )
         if interactive or devtools:
             use_existing_page = True
@@ -204,13 +249,17 @@ def _browser_context(
     devtools=False,
     retina=False,
     browser="chromium",
+    user_agent=None,
     timeout=None,
+    reduced_motion=False,
 ):
     browser_kwargs = dict(headless=not interactive, devtools=devtools)
     if browser == "chromium":
         browser_obj = p.chromium.launch(**browser_kwargs)
     elif browser == "firefox":
         browser_obj = p.firefox.launch(**browser_kwargs)
+    elif browser == "webkit":
+        browser_obj = p.webkit.launch(**browser_kwargs)
     else:
         browser_kwargs["channel"] = browser
         browser_obj = p.chromium.launch(**browser_kwargs)
@@ -219,6 +268,10 @@ def _browser_context(
         context_args["storage_state"] = json.load(auth)
     if retina:
         context_args["device_scale_factor"] = 2
+    if reduced_motion:
+        context_args["reduced_motion"] = "reduce"
+    if user_agent is not None:
+        context_args["user_agent"] = user_agent
     context = browser_obj.new_context(**context_args)
     if timeout:
         context.set_default_timeout(timeout)
@@ -240,8 +293,27 @@ def _browser_context(
     help="Wait this many milliseconds before failing",
 )
 @click.option("--fail-on-error", is_flag=True, help="Fail noisily on error")
+@click.option(
+    "noclobber",
+    "-n",
+    "--no-clobber",
+    is_flag=True,
+    help="Skip images that already exist",
+)
 @browser_option
-def multi(config, auth, retina, timeout, fail_on_error, browser):
+@user_agent_option
+@reduced_motion_option
+def multi(
+    config,
+    auth,
+    retina,
+    timeout,
+    fail_on_error,
+    noclobber,
+    browser,
+    user_agent,
+    reduced_motion,
+):
     """
     Take multiple screenshots, defined by a YAML file
 
@@ -254,6 +326,8 @@ def multi(config, auth, retina, timeout, fail_on_error, browser):
     \b
         - output: example.png
           url: http://www.example.com/
+
+    https://github.com/simonw/shot-scraper/blob/main/README.md#multi
     """
     shots = yaml.safe_load(config)
     if shots is None:
@@ -262,9 +336,21 @@ def multi(config, auth, retina, timeout, fail_on_error, browser):
         raise click.ClickException("YAML file must contain a list")
     with sync_playwright() as p:
         context, browser_obj = _browser_context(
-            p, auth, retina=retina, browser=browser, timeout=timeout
+            p,
+            auth,
+            retina=retina,
+            browser=browser,
+            user_agent=user_agent,
+            timeout=timeout,
+            reduced_motion=reduced_motion,
         )
         for shot in shots:
+            if (
+                noclobber
+                and shot.get("output")
+                and pathlib.Path(shot["output"]).exists()
+            ):
+                continue
             try:
                 take_shot(context, shot)
             except TimeoutError as e:
@@ -291,7 +377,12 @@ def multi(config, auth, retina, timeout, fail_on_error, browser):
     default="-",
 )
 @click.option("-j", "--javascript", help="Execute this JS prior to taking the snapshot")
-def accessibility(url, auth, output, javascript):
+@click.option(
+    "--timeout",
+    type=int,
+    help="Wait this many milliseconds before failing",
+)
+def accessibility(url, auth, output, javascript, timeout):
     """
     Dump the Chromium accessibility tree for the specifed page
 
@@ -301,7 +392,7 @@ def accessibility(url, auth, output, javascript):
     """
     url = url_or_file_path(url, _check_and_absolutize)
     with sync_playwright() as p:
-        context, browser_obj = _browser_context(p, auth)
+        context, browser_obj = _browser_context(p, auth, timeout=timeout)
         page = context.new_page()
         page.goto(url)
         if javascript:
@@ -336,7 +427,11 @@ def accessibility(url, auth, output, javascript):
     help="Save output JSON to this file",
 )
 @browser_option
-def javascript(url, javascript, input, auth, output, browser):
+@user_agent_option
+@reduced_motion_option
+def javascript(
+    url, javascript, input, auth, output, browser, user_agent, reduced_motion
+):
     """
     Execute JavaScript against the page and return the result as JSON
 
@@ -366,7 +461,13 @@ def javascript(url, javascript, input, auth, output, browser):
         javascript = input.read()
     url = url_or_file_path(url, _check_and_absolutize)
     with sync_playwright() as p:
-        context, browser_obj = _browser_context(p, auth, browser=browser)
+        context, browser_obj = _browser_context(
+            p,
+            auth,
+            browser=browser,
+            user_agent=user_agent,
+            reduced_motion=reduced_motion,
+        )
         page = context.new_page()
         page.goto(url)
         result = _evaluate_js(page, javascript)
@@ -406,7 +507,7 @@ def pdf(url, auth, output, javascript, wait, media_screen, landscape):
 
     Use -o to specify a filename:
 
-        shot-scarper pdf https://datasette.io/ -o datasette.pdf
+        shot-scraper pdf https://datasette.io/ -o datasette.pdf
     """
     url = url_or_file_path(url, _check_and_absolutize)
     if output is None:
@@ -471,7 +572,12 @@ def install(browser):
     "context_file",
     type=click.Path(file_okay=True, writable=True, dir_okay=False, allow_dash=True),
 )
-def auth(url, context_file):
+@browser_option
+@user_agent_option
+@click.option(
+    "--devtools", is_flag=True, help="Open browser DevTools"
+)
+def auth(url, context_file, browser, user_agent, devtools):
     """
     Open a browser so user can manually authenticate with the specified site,
     then save the resulting authentication context to a file.
@@ -481,8 +587,15 @@ def auth(url, context_file):
         shot-scraper auth https://github.com/ auth.json
     """
     with sync_playwright() as p:
-        browser = p.chromium.launch(headless=False)
-        context = browser.new_context()
+        context, browser_obj = _browser_context(
+            p,
+            auth=None,
+            interactive=True,
+            devtools=devtools,
+            browser=browser,
+            user_agent=user_agent,
+        )
+        context = browser_obj.new_context()
         page = context.new_page()
         page.goto(url)
         click.echo("Hit <enter> after you have signed in:", err=True)
@@ -528,10 +641,19 @@ def take_shot(
     wait = shot.get("wait")
     padding = shot.get("padding") or 0
 
-    # If a single 'selector' turn that into selectors array with one item
     selectors = shot.get("selectors") or []
+    selectors_all = shot.get("selectors_all") or []
+    js_selectors = shot.get("js_selectors") or []
+    js_selectors_all = shot.get("js_selectors_all") or []
+    # If a single 'selector' append to 'selectors' array (and 'js_selectors' etc)
     if shot.get("selector"):
         selectors.append(shot["selector"])
+    if shot.get("selector_all"):
+        selectors_all.append(shot["selector_all"])
+    if shot.get("js_selector"):
+        js_selectors.append(shot["js_selector"])
+    if shot.get("js_selector_all"):
+        js_selectors_all.append(shot["js_selector_all"])
 
     if not use_existing_page:
         page = context_or_page.new_page()
@@ -563,21 +685,45 @@ def take_shot(
     if not return_bytes:
         screenshot_args["path"] = output
 
-    if not selectors:
+    if (
+        not selectors
+        and not js_selectors
+        and not selectors_all
+        and not js_selectors_all
+    ):
         screenshot_args["full_page"] = full_page
 
-    if selectors:
-        # Use JavaScript to create a box around those elements
+    if js_selectors or js_selectors_all:
+        # Evaluate JavaScript adding classes we can select on
+        (
+            js_selector_javascript,
+            extra_selectors,
+            extra_selectors_all,
+        ) = _js_selector_javascript(js_selectors, js_selectors_all)
+        selectors.extend(extra_selectors)
+        selectors_all.extend(extra_selectors_all)
+        _evaluate_js(page, js_selector_javascript)
+
+    if selectors or selectors_all:
+        # Use JavaScript to create a box around those elementsdef
         selector_javascript, selector_to_shoot = _selector_javascript(
-            selectors, padding
+            selectors, selectors_all, padding
         )
         _evaluate_js(page, selector_javascript)
+        try:
+            bytes_ = page.locator(selector_to_shoot).screenshot(**screenshot_args)
+        except TimeoutError as e:
+            raise click.ClickException(
+                "Timed out while waiting for element to become available.\n\n{}".format(
+                    e
+                )
+            )
         if return_bytes:
-            return page.locator(selector_to_shoot).screenshot(**screenshot_args)
+            return bytes_
         else:
             page.locator(selector_to_shoot).screenshot(**screenshot_args)
             message = "Screenshot of '{}' on '{}' written to '{}'".format(
-                ", ".join(selectors), url, output
+                ", ".join(list(selectors) + list(selectors_all)), url, output
             )
     else:
         # Whole page
@@ -589,7 +735,43 @@ def take_shot(
     click.echo(message, err=True)
 
 
-def _selector_javascript(selectors, padding=0):
+def _js_selector_javascript(js_selectors, js_selectors_all):
+    extra_selectors = []
+    extra_selectors_all = []
+    js_blocks = []
+    for js_selector in js_selectors:
+        klass = "js-selector-{}".format(secrets.token_hex(16))
+        extra_selectors.append(".{}".format(klass))
+        js_blocks.append(
+            textwrap.dedent(
+                """
+        Array.from(
+          document.getElementsByTagName('*')
+        ).find(el => {}).classList.add("{}");
+        """.format(
+                    js_selector, klass
+                )
+            )
+        )
+    for js_selector_all in js_selectors_all:
+        klass = "js-selector-all-{}".format(secrets.token_hex(16))
+        extra_selectors_all.append(".{}".format(klass))
+        js_blocks.append(
+            textwrap.dedent(
+                """
+        Array.from(
+          document.getElementsByTagName('*')
+        ).filter(el => {}).forEach(el => el.classList.add("{}"));
+        """.format(
+                    js_selector_all, klass
+                )
+            )
+        )
+    js_selector_javascript = "() => {" + "\n".join(js_blocks) + "}"
+    return js_selector_javascript, extra_selectors, extra_selectors_all
+
+
+def _selector_javascript(selectors, selectors_all, padding=0):
     selector_to_shoot = "shot-scraper-{}".format(secrets.token_hex(8))
     selector_javascript = textwrap.dedent(
         """
@@ -600,6 +782,8 @@ def _selector_javascript(selectors, padding=0):
         let maxBottom = 0;
         let maxRight = 0;
         let els = %s.map(s => document.querySelector(s));
+        // Add the --selector-all elements
+        %s.map(s => els.push(...document.querySelectorAll(s)));
         els.forEach(el => {
             let rect = el.getBoundingClientRect();
             if (rect.top < minTop) {
@@ -638,7 +822,12 @@ def _selector_javascript(selectors, padding=0):
         }, 300);
     });
     """
-        % (padding, json.dumps(selectors), json.dumps(selector_to_shoot))
+        % (
+            padding,
+            json.dumps(selectors),
+            json.dumps(selectors_all),
+            json.dumps(selector_to_shoot),
+        )
     )
     return selector_javascript, "#" + selector_to_shoot
 
